package aql

import (
	"encoding/json"
	"fmt"
	"strconv"
)

// Constraint datatypes
// text		  MatchTypes: exact/contains/startswith/endswith
// number   MatchTypes: GT/LT/EQ
// bool     MatchTypes: EQ/NEQ

// Ranges dus tussen 10 half 5 bijv.

// Struct used for JSON conversion
type parsedJSON struct {
	Return    returnStruct
	Nodes     []nodeStruct
	Relations []relationStruct
}

type returnStruct struct {
	Entities  []int
	Relations []int
}

type nodeStruct struct {
	NodeType    string
	Constraints map[string]constraintStruct
}
type relationStruct struct {
	RelationType string
	NodeFrom     int
	NodeTo       int
	Depth        searchDepthStruct
	Constraints  map[string]constraintStruct
}
type searchDepthStruct struct {
	Min int
	Max int
}

type constraintStruct struct {
	Value     string
	DataType  string
	MatchType string
}

// ConvertJSONToAQL converts a json string to an AQL query
func ConvertJSONToAQL(jsonMsg *[]byte) (*string, error) {

	jsonStruct, err := convertJSONToStruct(jsonMsg)
	if err != nil {
		fmt.Println(err)
		return nil, err
	}

	//Per node query
	//	per constraint
	//relations koppelen ze samen
	//return statement

	result := createAllNodesQuery(jsonStruct.Return.Entities, jsonStruct.Nodes)
	return result, nil
}

func createAllNodesQuery(returnEntitiesIndices []int, nodes []nodeStruct) *string {
	var result string
	for nodeIndex := range returnEntitiesIndices {
		nodeID := fmt.Sprintf("n%s", strconv.Itoa(nodeIndex))

		nodeQueryString := *createNodeQuery(&nodes[nodeIndex], nodeID)

		result += fmt.Sprintf(" \n%s", nodeQueryString)
	}

	return &result
}

// createNodeQuery converts the node part of the json to a subquery
func createNodeQuery(node *nodeStruct, name string) *string {
	/*
		LET alices = (
			FOR x IN female
			FILTER x.name == "Alice" AND x.birth_year > 1997
			RETURN x
		)

		NAAR -->

		LET {NAAM**} = (
			FOR x IN {NODETYPE}
			FILTER x.{CONSTRAINT[0]} {{CONSTRAINT[0]}.MATCHTYPE} {CONSTRAINT[0].VALUE}
			AND x.{CONSTRAINT[1]} {{CONSTRAINT[1]}.MATCHTYPE} {CONSTRAINT[1].VALUE}
			RETURN x
		)

	*/

	letStatement := fmt.Sprintf("LET %s = (\n", name)
	forStatement := fmt.Sprintf("\tFOR x IN %s \n", node.NodeType)

	// Generate all constraints as FILTER statements
	first := true
	filter := "\tFILTER "
	for key, constraint := range node.Constraints {
		constraint := createQueryConstraint(&constraint, key)

		if first {
			filter += fmt.Sprintf("\t%s ", *constraint)
			first = false
		} else {
			filter += fmt.Sprintf("AND\n\t\t%s", *constraint)
		}
	}

	returnStatement := "\n\tRETURN x\n)"

	// Concatenate all the statements
	result := letStatement + forStatement + filter + returnStatement
	return &result
}

// Constraint datatypes
// text		  MatchTypes: exact/contains/startswith/endswith
// number   MatchTypes: GT/LT/EQ
// bool     MatchTypes: EQ/NEQ

// createQueryConstraint creates a sinlge line of AQL filtering/constraint
func createQueryConstraint(con *constraintStruct, key string) *string {
	//FILTER x.{CONSTRAINT[0]} {{CONSTRAINT[0]}.MATCHTYPE} {CONSTRAINT[0].VALUE}
	//			name				mtch					val + dataType
	var (
		mtch string
		val  string
		line string
	)

	//Wicked switches letsgo
	switch con.DataType {
	case "text":
		val = fmt.Sprintf("\"%s\"", con.Value)
		switch con.MatchType {
		case "contains":
			mtch = "IN"
		case "startswith":
			mtch = "LIKE"
			val = fmt.Sprintf("\"%s%%\"", con.Value)
		case "endswith":
			mtch = "LIKE"
			val = fmt.Sprintf("\"_%s\"", con.Value)
		default: //exact
			mtch = "=="
		}
	case "number":
		val = con.Value
		switch con.MatchType {
		case "GT":
			mtch = ">"
		case "LT":
			mtch = "<"
		case "GET":
			mtch = ">="
		case "LET":
			mtch = "<="
		default: //EQ
			mtch = "=="
		}
	default: /*bool*/
		val = con.Value
		switch con.MatchType {
		case "NEQ":
			mtch = "!="
		default: //EQ
			mtch = "=="
		}
	}
	line = fmt.Sprintf("x.%s %s %s", key, mtch, val)
	return &line
}

func convertJSONToStruct(jsonMsg *[]byte) (*parsedJSON, error) {
	jsonStruct := parsedJSON{}
	err := json.Unmarshal([]byte(*jsonMsg), &jsonStruct)

	if err != nil {
		return nil, err
	}

	return &jsonStruct, nil
}

/*
	desired output

	WITH male
	FOR a1 IN female
		FILTER a1.name == "Alice"

		(FOR r1v,r1e,r1p IN 1..1 OUTBOUND a1 relation
			FILTER r1v.name == "Bob")
			OR
		(FOR r2v,r2e,r2p IN 1..1 OUTBOUND a1 relation
			FILTER r2v.name == "Martin" && r2v.hasdog == true)
																// constraints for Bob
				OR r1v.name == "Martin"
			FILTER r1e.type == "married"

			FOR r2v,r2e,r2p IN 1..1 OUTBOUND r1v relation
				FILTER r2v.name == "Figo"
				FILTER r2e.type == "has_dog"

				RETURN {a1,r1v,r2v}


	of


	LET alices = (
		FOR x IN female
			FILTER x.name == "Alice"
			RETURN x
	)

	LET bobs = (
		FOR x IN male
			FILTER x.name == "Bob"
			RETURN x
	)

	LET alices = (
		FOR alice IN alices
			FOR r1v,r1e,r1p IN 1..1 OUTBOUND alice relation
				FILTER r1v IN bobs AND r1e.type == "married"
				RETURN {alice}
	)

	LET alices = (
		FOR alice IN alices
			FOR r1v,r1e,r1p IN 1..1 OUTBOUND alice relation
				FILTER r1v IN marting AND r1e.type == "friend"
				RETURN {alice}
	)

		FOR a2 IN male
			FILTER a2.name == "Bob"

			FOR r1v,r1e,r1p IN 1..1 OUTBOUND a1 relation
				FILTER r1v._id == a2._id
				FILTER r1e.type == "married"
				RETURN {a1,a2,r1e}

*/
/*
{
	"NodeType": "female",
	"Constraints":
		{
			"name":       { "Value": "Alice", "DataType": "text",   "MatchType": "exact" },
			"birth_year": { "Value": "1997",  "DataType": "number", "MatchType": "GT"    }
		}
	}

	NAAR -->

	LET alices = (
		FOR x IN female
			FILTER x.name == "Alice" AND x.birth_year > 1997
			RETURN x
	)

*/

//Nog een manier vinden om namen te syncen over de queries heen **

// func forGlory() *constraintStruct {
// 	var yeet constraintStruct
// 	yeet = constraintStruct{
// 		ConstraintName: "name",
// 		Value:          "Alice",
// 		MatchType:      "exact",
// 		DataType:       "text",
// 	}
// 	return &yeet
// }

// func alsoForGlory() {
// 	con := forGlory()

// 	toPrint := createQueryConstraint(*con)
// 	fmt.Println(*toPrint)
// }
